% KADMOS: a Finite Element code for the calculation of apparent K-Ar ages in minerals
% Evangelos Moulas and Mark T Brandon
% doi: 10.5281/zenodo.7358138
% 
% code written by Evangelos Moulas, 30/09/2022
% Metamorphic Processes Research Group
% Johannes Gutenberg University, Mainz, Germany
%========================================================================
clear,clc,close all
%Problem Set UP =========================================================
tic
Inputs_Path;
%Flags ==================================================================
flag_save   = 0;                 %Indicate if you want to save
flag_plot   = 1;                 %Indicate if you want to plot
flag_echo   = 1;                 %Indicate if you want to echo the results
%Check Number of Domains -------------------------------------------------
[ndd]                   = check_ndomains(Pres_a,TempK_a,time_a,Lr,ndim_a,Age0,Min_Sys);
%Numerical Parameters for FEM solution ==================================
chol_sol                =         1;           %Cholesky Solver
theta                   =         1;           %Crank-Nicolson Parameter
nt                      =       100;           %Time increments
nels                    =       100;           %Number of Elements
RdX                     =         2;           %Ratio of dX (1 for regular grid)
%plot options --------------------------------------------------------
nout                    =         2;           %Plot every nout iterations
LWW                     =       1.2;           %Linewidth of lines (in plot)
FSS                     =        12;           %Fontsize (in plot)
%========================================================================
%============== PLEASE DO NOT MODIFY BEYOND THIS POINT ==================
%Preprocessing & FEM ----------------------------------------------------
dt                      =   tlim/nt;           %Time discretization
nr                      =    nels+1;           %nx nodes in r-direction
[r1d,dr_ar,Cfact]       = linspace_adapt(0,1,nr,RdX);
%------------------------------------------------------------------------
el2n(1,:)    =    1:1:nr-1; %Extract nodes of element
el2n(2,:)    =    2:1:nr;   %nodes = el2n(:,elem);
%DOF-Node relation-------------------------------------------------------
el2d         =    el2n;
ndofs        =    nr*1;
%Usefull Constants ======================================================
c2K         =            273.15; %celcius 2 Kelvin
Myr2sec     =  3600*24*365.25e6; %Myr in secs
mic2met     =              1e-6; %from microns to cm
cm2metr     =              1e-2; %from cm to metre
micr2cm     =   mic2met/cm2metr; %from microns to cm
Rgas        =         8.3144621; %gas constant
TsurfK      = 25+c2K;            %Tsurface
TminK       = Tmin + c2K;        %MinT of Interpolation
f_loss      = 1.1247;            %Critical fractional loss value
%Argon decay----------------------------------------------------------
lamti       = 5.543*1e10*1e-6;   %Inverse tot Lamda in Myr
lam1i       = 0.581*1e10*1e-6;   %Inverse for lamda in Myr (no Ca)
%Preprocess for decay constants---------------------------------------
lamt        =        1./lamti;   %decay constants for Ar
lam1        =        1./lam1i;   %-//-
ke          =       lam1/lamt;   %ratio to total
%Initial Conditions --------------------------------------------------
rplot           =  repmat(r1d(:),1,ndd);         %Ndim Radial dist
Cloc            =  (1/ke).*(1-exp(-lamt.*Age0)); %Initial C
Cplot           =  repmat(Cloc,nr,1);            %Conc profiles
Cplot(end,:)    =  0;
C               =  Cplot(:);
t_ArK0          = -1/lamt*log(1-ke*Cplot); %Calculation of K-Ar Age
%Timeloop ------------------------------------------------------------
t_Myr = 0;
it    = 0;
while t_Myr<tlim
    it    = it+1;
    t_Myr = t_Myr+dt;
    %Time Change of Source Term & Diffusion Coefficient------------------
    Sn    =      1/ke*(lamt*exp(-lamt*(t_Myr   )));
    So    =      1/ke*(lamt*exp(-lamt*(t_Myr-dt)));
    %Integrate Source Separately =========================================
    C                =  C  +  dt*(theta*Sn+(1-theta)*So);
    C(nr:nr:nr*ndd)  =  0;
    %Diffusion Solution ==================================================
    Co    =    C(:);    %Store Old Concentration
    L_G   =      [];
    R_G   =      [];
    for idd = 1:ndd
        x_1   =    [r1d(1:end-1)]';
        x_2   =    [r1d(2:end-0)]';
        %Interp properties on path ---------------------------------
        Tn                   =       interp1(time_a(:,idd),TempK_a(:,idd),t_Myr,'linear',TminK);
        Pn                   =       interp1(time_a(:,idd),Pres_a(:,idd) ,t_Myr,'linear',Pmin);
        To                   =       interp1(time_a(:,idd),TempK_a(:,idd),t_Myr-dt,'linear',TminK);
        Po                   =       interp1(time_a(:,idd),Pres_a(:,idd) ,t_Myr-dt,'linear',Pmin);
        Trec(it,idd)         =       Tn-c2K;
        Prec(it,idd)         =       Pn;
        trec(it,idd)         =       t_Myr;
        [D0_cm2s,Ea,Va,Pref] =       diff_system(Min_Sys(idd));
        D0a2_myr             =       (D0_cm2s/(Lr(idd)*micr2cm)^2)*Myr2sec;
        Dn                   =       D0a2_myr*exp(-(Ea+(Pn-Pref)*Va)./Rgas./Tn);
        Do                   =       D0a2_myr*exp(-(Ea+(Po-Pref)*Va)./Rgas./To);
        ndim                 =       ndim_a(idd);
        D_reset(idd)         =       Dn*dt;
         if D_reset(idd)>f_loss;
             Dn = 0;
         end
        %-----------------------------------------------------------
        [L_g,R_g]            =       fill_MatrixFEM(idd,nr,ndofs,ndim,x_1,x_2,Dn,Do,theta,dt,Co);
        %Add block =======================================================
        L_G = blkdiag(L_G,L_g);
        R_G = [R_G;R_g];
    end    
    %Calculate Solution ==================================================
    if chol_sol == 1
        CH_g              = chol(L_G);        %Cholesky Factorization
        C                 = CH_g\(CH_g'\R_G); %Cholesky Solution
    else
        C                 = L_G\R_G;          %Solution
    end
    %Check Reset =========================================================
    ind_reset = zeros(ndd*nr,1);
    for idd = 1:ndd
        if D_reset(idd)>f_loss;
            C((idd-1)*nr+1:(idd-1)*nr+nr-1) = 0;
        end
    end
    %=====================================================================
    if mod(it,nout)==1 
        if flag_plot == 1
            Cplot = reshape(C,nr,ndd);
            t_ArK          = -1/lamt*log(1-ke*Cplot);  %Calculation of K-Ar Age
            %----------------------------------------------------
            subplot(121)
            plot(time_a-tlim,TempK_a-c2K,time_a-tlim,TempK_a-c2K,'LineWidth',LWW,'LineWidth',LWW)
            hold on
            plot(trec(1:nout:it,:)-tlim,Trec(1:nout:it,:),'k .','LineWidth',LWW)
            hold off
            grid on
            axis([-tlim 0.1 0 T0])
            xlabel('Age (Myr)','Interpreter','Latex','FontSize',FSS)
            ylabel('T ($^o$C)','Interpreter','Latex','FontSize',FSS)
            set(gca,'LineWidth',LWW,'FontSize',FSS)
            axis square
            
            subplot(122)
            plot(rplot,t_ArK,'LineWidth',LWW)
            grid on
            xlim([0,1.1])
            xlabel('radius [-]','Interpreter','Latex','FontSize',FSS)
            ylabel('Age (Myr)','Interpreter','Latex','FontSize',FSS)
            set(gca,'LineWidth',LWW,'FontSize',FSS)
            axis square
            title(['Max age : ',num2str(max(t_ArK(:)))])
            %---------------------------------------------------
            drawnow
        end
        if flag_echo == 1
            disp([num2str(round(it/nt*100),3),' % is done'])
        end
    end    
end
%====================================================================
time_total     = toc;
Cplot          = reshape(C,nr,ndd);
t_ArK          = -1/lamt*log(1-ke*Cplot);  %Calculation of K-Ar Age
%Parameters to save -------------------------------------------------
data_time_tot    = time_total;
data_Trec_C      = Trec;
data_trec_Myr    = trec;
data_agePath_Myr = trec-tlim;
data_C_Ar_K      = Cplot;
data_Ar_Age_Myr  = t_ArK;
data_L_microns   = Lr;
data_ndim_geom   = ndim_a;
data_Age0        = Age0;
data_Min_Sys     = Min_Sys;
data_ndd         = ndd;
data_name        = name_file;
data_r1d         = r1d;
data_nels        = nels;
data_nr          = nr;
data_El2n        = el2n;
data_El2d        = el2d;
data_ndofs       = ndofs;
data_rplot       = rplot;
data_Core_Ages   = t_ArK(1,:);
%=======================================================================
if flag_echo == 1
    disp(['Calculation is done'])
    disp(['Age (core) : ',num2str(data_Core_Ages),' Myr'])
    disp(['time for solution (sec): ',num2str(time_total)]);
end
if flag_plot == 1
    subplot(121)
    plot(time_a-tlim,TempK_a-c2K,time_a-tlim,TempK_a-c2K,'LineWidth',LWW)
    hold on
    plot(trec(1:nout:it,:)-tlim,Trec(1:nout:it,:),'k .','LineWidth',LWW)
    hold off
    grid on
    axis([-tlim 0.1 0 T0])
    xlabel('Age (Myr)','Interpreter','Latex','FontSize',FSS)
    ylabel('T ($^o$C)','Interpreter','Latex','FontSize',FSS)
    set(gca,'LineWidth',LWW,'FontSize',FSS)
    axis square
    
    subplot(122)
    plot(rplot,t_ArK,'LineWidth',LWW)
    grid on
    xlim([0,1.1])
    ylim([0 55])
    xlabel('radius [-]','Interpreter','Latex','FontSize',FSS)
    ylabel('Age (Myr)','Interpreter','Latex','FontSize',FSS)
    set(gca,'LineWidth',LWW,'FontSize',FSS)
    title(['Max age : ',num2str(max(t_ArK(:)))])
    axis square
end
if flag_save == 1
    clearvars -except data_Trec_C data_trec_Myr data_agePath_Myr data_C_Ar_K ...
              data_Ar_Age_Myr  data_L_microns data_ndim_geom data_Age0 ...
              data_Min_Sys data_ndd data_name data_r1d data_nr data_El2n data_El2d ...
              data_ndofs data_rplot data_nels data_Core_Ages
    strinG_temp = ['save data_path_',data_name]; eval(strinG_temp);    
end
%===================================================================
function [ndd] = check_ndomains(Pres_a,TempK_a,time_a,Lr,ndim_a,Age0,Min_Sys)
    [ndd0]        = size(Lr,2);
    [npath1,ndd1] = size(Pres_a);
    [npath2,ndd2] = size(TempK_a);
    [npath3,ndd3] = size(time_a);
    [ndd4]        = size(ndim_a,2);
    [ndd5]        = size(Age0,2);
    [ndd6]        = size(Min_Sys,2);
    if (ndd1~=ndd0) | (ndd2~=ndd0) | (ndd3~=ndd0) | (ndd3~=ndd0)...
                    | (ndd4~=ndd0) | (ndd5~=ndd0) | (ndd6~=ndd0)
        disp('Number of domains does not match')
        error('check size of PTt array, Size, Geometry, Initial Age or System')
    end
    ndd = ndd1;
end
function [D0_cm2s,Ea,Va,Pref] = diff_system(name_sys)
kcal2kj = 4.1868;
%Available Systems (see below for references)
%Muscovite 1 & 2
%Hornblende 1 & 2
%Biotite 1 & 2 & 3 & 4
%Orthoclase
if strcmp(name_sys,'Muscovite1')==1
    %Data from Harrison et al. 2009 Geoch - Cosm Acta
    %fitted using sph. geometry and grain size range 24 (most) to 87 microns
    D0_cm2s     = 2.3;               %Pre-exponent factor
    Ea          = 263*1e3;           %Activation Energy in j/mol
    Va          = 14;                %Activation volume in cm3, P in MPa
    Pref        = 1000;              %Reference Pressure (MPa) for PV term
elseif strcmp(name_sys,'Muscovite2')==1
    %as in Lister and Baldwin 1999 (Tectonophysics)
    %used for slab geometry and half thickness of 5.9 microns
    %Original data from Robbins (1972, MSc Thesis, State .Univ. NY at Albany)
    D0_cm2s     = 3.35*1e-7;         %Pre-exponent factor
    Ea          = 41.8*kcal2kj*1e3;  %Activation Energy in j/mol
    Va          = 10;                %Activation volume in cm3, P in MPa
    Pref        = 1e-4;              %Reference pressure in MPa
elseif strcmp(name_sys,'Hornblende1')==1
    %Experimental Set from Harrison 1982 Contrib Mineral Petrol
    %This is only the experimental set and not the final
    %expression suggested by the author
    %Activation volume correction after Lister and Baldwin 1996
    %(used for 80 microns and spherical geometry)
    D0_cm2s     = 0.061;             %Pre-exponent factor
    Ea          = 66.1*kcal2kj*1e3;  %Activation Energy in j/mol
    Va          = 10;                %Activation volume in cm3, P in MPa
    Pref        = 1e-4;              %Reference Pressure (MPa) for PV term
elseif strcmp(name_sys,'Hornblende2')==1
    %Harrison 1982 Contrib Min Petrol (parameters as suggested by author)
    %Activation volume correction after Lister and Baldwin 1996
    %(used for 80 microns abd spherical geometry)
    D0_cm2s     = 0.024;             %Pre-exponent factor
    Ea          = 64.1*kcal2kj*1e3;  %Activation Energy in j/mol
    Va          = 10;                %Activation volume in cm3, P in MPa
    Pref        = 1e-1;              %Reference Pressure (MPa) for PV term
elseif strcmp(name_sys,'Biotite1')==1
    %Original data from Grove and Harrison 1996 (Am Mineral)
    %Data for Fe-Biotite (Xann = 0.71)
    %Activation volume is not considered
    D0_cm2s     = 0.40;              %Pre-exponent factor
    Ea          = 50.5*kcal2kj*1e3;  %Activation Energy
    Va          = 0;                 %Originally 0, Skipton et al. (2018) used 14 as in Harrison et al. (2009)
    Pref        = 1e-4;              %Reference Pressure (MPa) for PV term
elseif strcmp(name_sys,'Biotite2')==1
    %as used in Lister and Baldwin 1996
    %Original data from Harrison et al. 1985 (Geochim Cosmochim Acta)
    %Activation volume correction after Lister and Baldwin 1996
    %(used for 150 microns and cylindrical geometry)    
    D0_cm2s     = 0.077;             %Pre-exponent factor
    Ea          = 47*kcal2kj*1e3;    %Activation Energy
    Va          = 10;                %Originally 0, Skipton et al. (2018) used 14 as in Harrison et al. (2009)
    Pref        = 1e-4;              %Reference Pressure (MPa) for PV term
elseif strcmp(name_sys,'Biotite3')==1
    %Original data from Giletti 1974 (as reported in McDougal & Harrison 1999)
    %Data for Phlogopite (Xann = 0.04)
    %Activation volume correction after Lister and Baldwin 1996
    %(used for 150 microns and cylindrical geometry) 
    D0_cm2s     = 0.75;              %Pre-exponent factor
    Ea          = 57.9*kcal2kj*1e3;  %Activation Energy
    Va          = 10;                %Originally 0
    Pref        = 1e-4;              %Reference Pressure (MPa) for PV term    
elseif strcmp(name_sys,'Biotite4')==1
    %Original data from Grove and Harrison 1996 (Am Mineral)
    %Data for Cooma biotite (Xann = 0.54)
    D0_cm2s     = 0.075;             %Pre-exponent factor
    Ea          = 47.1*kcal2kj*1e3;  %Activation Energy
    Va          = 0;                 %not considered
    Pref        = 1e-4;              %Reference Pressure (MPa) for PV term
elseif strcmp(name_sys,'Orthoclase')==1
    %Original data from Foland 1993
    D0_cm2s     = 0.00982;           %Pre-exponent factor
    Ea          = 43.8*kcal2kj*1e3;  %Activation Energy
    Va          = 0;                 %not considered
    Pref        = 1e-4;              %Reference Pressure (MPa) for PV term
elseif strcmp(name_sys,'custom')==1
    load Custom_Diff.dat
    D0_cm2s     = Custom_Diff(1);    %Value in cm2/sec
    Ea          = Custom_Diff(2);    %in kj/mol
    Va          = Custom_Diff(3);    %in cm3
    Pref        = Custom_Diff(4);    %in MPa
else
    error('Unknown mineral system')
end
end
function [x2,dx_ar,Cfact] = linspace_adapt(L1,L2,nx,dX1_dXN)
%Function linspace_adapt creates unequal linspace for grid construction
%This function is based on the ulinspace function (explained below)
%The Inputs are
%L1     : beggining of Array
%L2     : end of Array
%nx     : size of Array
%dX1_dXN: Ratio of 1st to last dx
C        = (dX1_dXN).^(1/(1-nx));
%Ulinspace function ----------------------------------------------------
ndx      = nx-1;
LHS      = eye(ndx);
LHS(1,:) = 1;
for i = 1:ndx-1
    LHS(i+1,i) = -C;    
end
RHS      = zeros(ndx,1);
RHS(1)   = L2-L1;
dx_ar    = (LHS\RHS)';
x2       = [0 cumsum(dx_ar)];
x2       = x2+L1;
dx_ar    = diff(x2);
Cfact    = C;
end
function [L_g,R_g]=fill_MatrixFEM(ig,nr,ndofs,ndim,x_1,x_2,Dn,Do,theta,dt,Co)
if ndim ~=1 & ndim~=2 & ndim ~=3
    error('Dimension (ndim) ERROR')
end
M_11i = (2.*x_2.^ndim.*x_2.^2)./(ndim.*(x_1 - x_2).^2.*(ndim.^2 + 3.*ndim + 2)) - ...
    (x_1.^ndim.*(ndim.^2.*x_1.^2 - 2.*ndim.^2.*x_1.*x_2 + ndim.^2.*x_2.^2   + ...
    ndim.*x_1.^2 - 4.*ndim.*x_1.*x_2 + 3.*ndim.*x_2.^2 + 2.*x_2.^2))./(ndim.*(x_1 - x_2).^2.*(ndim.^2 + 3.*ndim + 2));
M_22i = -((x_1.^ndim.*x_1.^2)./ndim - (x_1.^2.*x_2.^ndim)./ndim - ...
    (2.*x_1.^ndim.*x_1.^2)./(ndim + 1) + (x_1.^ndim.*x_1.^2)./(ndim + 2) - ...
    (x_2.^ndim.*x_2.^2)./(ndim + 2) + (2.*x_1.*x_2.*x_2.^ndim)./(ndim + 1))./(x_1 - x_2).^2;
M_12i = -(ndim.*(x_1.^ndim.*x_1.^2 - x_2.^ndim.*x_2.^2 - x_1.*x_1.^ndim.*x_2 + x_1.*x_2.*x_2.^ndim) - ...
    2.*x_1.*x_1.^ndim.*x_2 + 2.*x_1.*x_2.*x_2.^ndim)./(ndim.*(x_1 - x_2).^2.*(ndim.^2 + 3.*ndim + 2));
K_11i = -((x_1.^ndim - x_2.^ndim))./(ndim.*(x_1 - x_2).^2);
K_12i =  ((x_1.^ndim - x_2.^ndim))./(ndim.*(x_1 - x_2).^2);
indE  = [1:length(x_1)]'; %Index as number of Elements
Nels  = indE(end);        %Number of Elements
Ndof  = Nels*1+1;         %Total Number of Dofs
M11d  = [M_11i;0];
M22d  = [0;M_22i];
M12d  = [0;M_12i];
K11d  = [K_11i;0];
K22d  = [0;K_11i];
K12d  = [0;K_12i];
Md_0  = M11d+M22d;
Md_1  = M12d;
Kd_0  = K11d+K22d;
Kd_1  = K12d;
M11m  = spdiags([Md_0],0,Ndof,Ndof);
M11s  = spdiags([Md_1],1,Ndof,Ndof);
K11m  = spdiags([Kd_0],0,Ndof,Ndof);
K11s  = spdiags([Kd_1],1,Ndof,Ndof);
%Derivatives of Last Shape Function
xE    = x_2(end);
dN1   = -1/(x_2(end)-x_1(end));
dN2   = +1/(x_2(end)-x_1(end));
dNdx  = [dN1 dN2];
MM    = M11m+M11s+M11s';
KM    = K11m+K11s+K11s';
L_g   =  MM./dt +KM.*Dn.*(  theta);
R_g   = (MM./dt -KM.*Do.*(1-theta))*Co(1+nr*(ig-1):1+nr*(ig-1)+nr-1);
%Boundary Conditions only at end (Enforce symmetry for specific BC)
L_g(nr,:)         = 0;
L_g(:,nr)         = 0;
L_g(nr,nr)        = 1;
R_g(nr)           = 0;
end
